#include <stdio.h>
#include "loft_c.h"

extern void add_to_broadcast (scheduler_t,broadcast_t);
extern int direct_access;

/***************************************************/
event_t event_create_in (scheduler_t sched)
{
   event_t res    = mymalloc (sizeof (struct _event_t));
   res->max       = EVENT_VALUES_CHUNK;
   res->values    = mymalloc (res->max * sizeof (void*));
   res->length    = 0;
   res->instant   = -1;
   res->scheduler = sched;
   res->waiting   = _make_item_list ();
   return res;
}

event_t event_create (void)
{
   return event_create_in (implicit_scheduler ());
}

void event_deallocate (event_t event)
{
   myfree               (event->values);
   _deallocate_item_list (event->waiting);
   myfree               (event);
}

/***************************************************/
static void reset_event_values (event_t event)
{
   event->length = 0;
}

static void add_event_value (event_t event,void *val)
{
   if (event->length >= event->max) {
      event->max += EVENT_VALUES_CHUNK;
      event->values =
	 myrealloc (event->values,event->max * sizeof (void*));
   }
   event->values[event->length] = val;
   event->length++;
}

int event_values_length (event_t event)
{
   return event->length;
}

void *nth_of_event_values (event_t event,int k)
{
   return event->values[k];
}


/*********************************************/
void register_in_event (event_t event)
{
   item_cell_t cell =
      _append_item_list (event->waiting,event->scheduler->current_cell);
   self ()->awaited_event = event;
   self ()->event_cell = cell;
}

/*********************************************/
static int awake_one (void *item,void *arg)
{
   item_cell_t cell = item;
   thread_t thread = _content_of_cell (cell);
   scheduler_t sched = arg;
   if (thread->state == _WAIT) {
      unmask (sched,cell);
      unregister_in_timer (thread);
   }
   return REMOVE;
}

static void awake (event_t event)
{
   _apply_on_items (event->waiting,awake_one,event->scheduler);
}

/***************************************************/
int is_present (event_t event)
{
   return event->scheduler->instant == event->instant;
}

void broadcast (event_t event)
{
   if (!is_present (event)) {
      scheduler_t sched = event->scheduler;
      reset_event_values (event);
      event->instant = sched->instant;
      sched->move = 1;
      awake (event);
   }
}

void broadcast_value (event_t event,void *val)
{
   broadcast (event);
   add_event_value (event,val);
   event->scheduler->move = 1;
}

/************************************************************/
void generate (event_t event)
{
   scheduler_t sched = event->scheduler;
   if (current_scheduler () != sched) {
      add_to_broadcast (sched,make_broadcast (event,1,NULL));
   } else {
      broadcast (event);
   }
}
 
void generate_value (event_t event,void *val)
{
   scheduler_t sched = event->scheduler;
   if (current_scheduler () != sched) {
      add_to_broadcast (sched,make_broadcast (event,0,val));
   } else {
      broadcast_value (event,val);
   }
}

/************************************************************/
