#include "loft_c.h"

thread_t make_thread (scheduler_t sched,
		      automaton_t autom,
		      finalizer_t final,
		      void *locals)
{
   thread_t res        = mymalloc (sizeof (struct _thread_t));
   res->term_signal    = event_create_in (sched);
   res->automaton      = autom;
   res->finalizer      = final;   
   res->state          = _CONT; // initially ready to run
   res->suspended      = 0;
   res->called         = NULL;
   res->code           = OK;
   res->locals         = locals;
   res->scheduler      = sched;

   res->awaited_event  = NULL;
   res->event_cell     = NULL;
   res->timer_cell     = NULL;

   res->magic          = MAGIC;
   return res;
}

/****************************************************/
void deallocate_thread (thread_t thread)
{
   myfree             (thread->locals);
   event_deallocate   (thread->term_signal);
   thread->magic = 0; // the thread does not exist anymore
   myfree             (thread);
}

/****************************************************/
int fireable (thread_t thread)
{
   return (!thread->suspended) && (_CONT == thread->state);
}

/****************************************************/
static void return_processing (thread_t thread,char val)
{
   thread->state = (_RTRN == val) ? _TERM : val;
}

/****************************************************/
void normal_termination (thread_t thread)
{
   thread->state = _TERM;
   generate (thread->term_signal);
}

/****************************************************/
void finalize_thread (thread_t thread)
{
   normal_termination (thread);
   thread->finalizer (thread->locals);
   if (thread->called) finalize_thread (thread->called);
}

/****************************************************/
thread_t self (void)
{
   return current_scheduler ()->current;
}

/****************************************************/
void fire_thread (thread_t thread)
{
   thread->scheduler->current = thread; // set self
   return_processing (thread,thread->automaton (thread->locals));
}

/****************************************************/
int return_code (void)
{
   return self ()->code;
}

/****************************************************/
int reset_thread_state (thread_t thread)
{
   if (!thread->suspended) {
      if (_COOP == thread->state) thread->state = _CONT;
      if (_CONT == thread->state) return 1;
   }
   return 0;
}

/****************************************************/
static void reinit_thread (thread_t thread)
{
   thread->timer_cell = NULL;
   thread->awaited_event = NULL;
   thread->event_cell = NULL;
}

void unregister_in_timer (thread_t thread)
{
   item_cell_t cell = thread->timer_cell;
   if (cell) {
      _extract (thread->scheduler->timer,cell);
      myfree (cell);
   }
   reinit_thread (thread);
}

void unregister_in_event (thread_t thread)
{
   item_cell_t cell = thread->event_cell;
   event_t event = thread->awaited_event;
   if (cell && event) {
      _extract (event->waiting,cell);
      myfree (cell);
   }
   reinit_thread (thread);
}

/****************************************************/
void* local_data (thread_t thread)
{
   return thread->locals;
}

/****************************************************/
